#ifndef CELL_LIST_H
#define CELL_LIST_H

/*This library contains a set of fuctions for dealing with cell lists, in molecular
 *simulations. The cell list must be a cube, and is centred at the origin of the (catesian) frame
 * that the simulation is in.
 *
 * the module prefic is CLL_
 */


/*a struct containing all the cell list data
 *
 * a list containing the id of the sell each particle is in
 * head_of: a list containing the particle at the head of the linked list for each cell
 * ll: the linked list for the cells
 * **neighbours: a list of the neighbouring cells of each cell (cells are not their own neighbour!)
 * double size: the width of the cell
 * int number: the number of cells on a side
 */

typedef struct clmc {

	long *of_particle;
	long *head_of;
	long *ll;
	long **neighbours;
	double size;
	long number;

} Mc_cell;

/*
 * A function for initialiseing the cell struct. It fills in the neighbours list,
 * the nunmber of cells on a side and the size of the cell.
 *
 * parameter Mc_cell cell: The struct containing all the cell list information
 * double boxS: the size of the simulation cube.
 * double cell_min is the minimum cell size (ie. the cut-off distance of the potential)
 */

void CLL_init(Mc_cell *cell, double boxS, double cell_min);

/*perform a deepcopy of cellI into cellT
 *
 *param cellT the target cell to be copies into. Current contents are overwritten
 *input cellI the cell list to be used as the source of the copy (not editied)
 *n_new, the new number of particles to go into cellT
 *n old the old number of particlesint cellT
 */


void CLL_cell_deepCopy (Mc_cell *cellT, Mc_cell *cellI, int n_new, int n_old);


/*clones cellI into the (uninitialised struct) cellT
 *param cellT the target cell to be copies into. Current contents are overwritten
 *input cellI the cell list to be used as the source of the copy (not editied)
 *n, the new number of particles in cellI
*/
void CLL_cell_clone (Mc_cell *cellT, Mc_cell *cellI, int n);

/*
 * A fuction for putting the initial group of particles into the cell list correctly
 *
 * param MC_cell cell
 * input Mc_state state: the stare struct of the particles in the system
 */

void CLL_init_particles(Mc_cell *cell, Mc_state *state);

/*
 * A function for updating the cell of a single particle
 *
 * param Mc_cell: the cell ist to be updated
 * input Mc_state: the current state of the system
 * input int: the index of the particle to be updated
 */

void CLL_update_particle(Mc_cell *cell, Mc_state *state, int index);


/*a function for adding a particle to the cell list. Assumes it is already in Mc_state
 * ie. call this function after adding the particle
 *
 * A function for updating the cell of a single particle
 *
 * param Mc_cell: the cell ist to be updated
 * input Mc_state: the current state of the system
 * input int: the index of the particle to be updated
 *
 **/

void CLL_add_particle(Mc_cell *cell, Mc_state *state, int index);

/*a function for removing a particle from the cell list. Assumes it is still in MC_state
 * ie. call before removing the particle from the state struct!!!
 * A function for updating the cell of a single particle
 *
 * param Mc_cell: the cell ist to be updated
 * input Mc_state: the current state of the system
 * input int: the index of the particle to be updated
 *
 **/

void CLL_remove_particle(Mc_cell *cell, Mc_state *state, int index);



/* a Fuction to get the index of the cell that an imaginary particle at u,v would be in. Does not 
 * edit the cell list.
 *
 * Output: the cell index
 * input: u,v, the generalised coords of u and v
 */

long CLL_get_index(double x, double y, double z, Mc_cell *cell);

/*destroys a cell struct*/
void CLL_end(Mc_cell *cell);


#endif
