/*This is a module containing ALL the code related to doing cluster moves,
 * using data augmentation on a sphere.
 *
 * The prefix is CMV_
 */

/*struct clusters: This structure contains all the information about the current partition of 
 *particles in a system.
 *
 * int head_of[] a list containing the index of the particle at the head
 * of the cluster for each particle
 * ll[]: a linked list. the value at the index of a given particle
 * int cluster_of[]: the index of the cluster of each particle
 */

#include"mc_math_h.h"
#include"std_h.h"
#include"mc_state_h.h"
#include"cell_list_h.h"
#include"cluster_moves_h.h"
#include"twister_h.h"
#include"geometry_h.h"

/*-----------------------------------helper function------------------------------*/
void CMV_connect (Mc_state *state, Mc_cell *cells, Mc_clusters *cluster,  
			int i, int j, double psudo_temp, int periodic, double *paramsG,
			double (*potential)(double,double*), double paramsP[]) ;
double CMV_energy_calc (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster,
		        double x, double y, double z, int periodic, double *paramsG,
			int index, int cell_of, double psudo_temp,
			double (*potential)(double,double*), double paramsP[]);

/*------------------------------------public functions--------------------------------------*/

/*FUNCTION: CMV_init allocates the memeory for the clusters
 *
 * param cluster: cluster struct to be initialised
 * inoput n_particles: the total number of particles in the system
 */

void CMV_init(Mc_clusters *cluster, int n_particles) {
	cluster->head_of = malloc(sizeof(int)*n_particles);
	cluster->ll = malloc(sizeof(int)*n_particles);

}

/*FUNTION: CMV_destroy: frees the structure*/

void CMV_destroy(Mc_clusters *cluster) {
	free(cluster->head_of);
	free(cluster->ll);
}

/*Function CMV_get psudo temperature: returns a psudo_temoerature between max and min
 *
 *input max: the mxaimum allwoed psudotemperature
 *input min: the minimum allowed psudo_temperature
 */

double CMV_get_psudo_temp (double max, double min) {
	return (max-min)*randomF() + min;
}


/*Function: Get clusters: This function partitions a system of N particles on a sphere into 
*	clusters, in accordance to the algorithm.
*
* 	Input: Mc_state: the current state of the system
*	Input: Mc_cell: the cell list for the system
*	output: Mc_clusters: the new cluster partition 
*/

void CMV_get_clusters (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster, double psudo_temp,
			int periodic, double *paramsG,
			double (*potential)(double,double*), double paramsP[]) {

	/*some dummies*/
	int i,j,k,n,current_cell,cell_of;
	/*initialise everything to zero*/
	/*Every particle is in a cluster contining only itself*/

	for (i = 0; i < state->n_particles; i++) {
		cluster->head_of[i] = i;
		cluster->ll[i] = -1;
	}
	/*handle special case of only one particle*/
	if (state->n_particles == 1) {
		cluster->head_of[0] = 0;
		cluster->ll[0] = -1;;
	} else {

		for (i = 0; i < state->n_particles; i++) {
			cell_of = cell->of_particle[i];
			n = cell->head_of[cell_of];
			while (n != i && i != -1) {
				if (n > i) CMV_connect(state,cell,cluster,i,n,psudo_temp,
							periodic, paramsG, potential, paramsP);
				n = cell->ll[n];	
			}
			if (n != -1) {
				n = cell->ll[n];
			}
			while (n != -1) {
				if (n > i) CMV_connect(state,cell,cluster,i,n,psudo_temp,
							periodic, paramsG, potential, paramsP);
				n = cell->ll[n];	
			}
			for (j = 0; j < 26; j++) {
				if (cell->neighbours[cell_of][j] == -1) break;
				current_cell = cell->neighbours[cell_of][j];
				n = cell->head_of[current_cell];
				while (n != -1) {
					if (n > i) CMV_connect(state,cell,cluster,i,n,psudo_temp,
							periodic, paramsG, potential, paramsP);
					n = cell->ll[n];	
				}
				
			}
		}
	}
	
}

/* FUNCTION: move_mc takes a whole cluster and makes a monte carlo move suggesting 
 * it's movement . The new
 * position of the centre of the mass of the cluster moves arounnd the sphere
 * by an angle less than delta_cluster.
*  Input head: the head of the cluster to be moved
* 	Input: Mc_state: the current state of the system
*	Input: Mc_cell: the cell list for the system
*	Input MC_cluster: the clustering struct
*	Input delta: the max length of the move (an radians aroud the sphere)
*	Input: radius: the radius of the sphere
*	Input: move_cluster: Function that moves a cluster in the current geometry
*/

int CMV_move_mc (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster, double delta,
			double *radius, double psudo_temp, int periodic, double *paramsG,
			double (*potential)(double,double*), 
			void (*move_cluster)(Mc_state*,Mc_cell*,Mc_clusters*,int,double,double*),
			double paramsP[]) {

	Mc_state *old_state = malloc(sizeof(Mc_state));
	Mc_cell *old_cell   = malloc(sizeof(Mc_cell));
	MCS_state_clone(old_state,state);
	CLL_cell_clone(old_cell,cell,state->n_particles);
	/*select cluster to move*/
	int head = cluster->head_of[randomMT() % state->n_particles];

	double old_energy = 0;
	int n = head;
	while ( n != -1) {
		old_energy += CMV_energy_calc(state,cell,cluster, state->x[n], state->y[n], 
					      state->z[n], periodic, paramsG, n, 
					      cell->of_particle[n], psudo_temp,potential,paramsP);
		n = cluster->ll[n];
	}
	move_cluster(state,cell,cluster,head,delta,radius);
	double energy = 0;

	n = head;
	while ( n != -1) {
		energy += CMV_energy_calc(state,cell,cluster, state->x[n], state->y[n], 
					      state->z[n], periodic, paramsG, n, 
					      cell->of_particle[n], psudo_temp,potential,paramsP);
		n = cluster->ll[n];
	}	
	
	//accepts_reject
	if (randomF() < exp (-1 * (energy-old_energy))) {
		CLL_end(old_cell);
		MCS_end(old_state);
		free(old_cell);
		free(old_state);
		return 1;
	} else {
		MCS_state_deepCopy(state,old_state);
		CLL_cell_deepCopy(cell, old_cell,state->n_particles,state->n_particles);
		CLL_end(old_cell);
		MCS_end(old_state);
		free(old_cell);
		free(old_state);
		return 0;
	}


}

/* FUNCTION: spin_mc takes a whole cluster and suggests a monte carlo move that spins it as a whole
*  spins it as a whole. The centre of mass does not change.
*  Input head: the head of the cluster to be moved
* 	Input: Mc_state: the current state of the system
*	Input: Mc_cell: the cell list for the system
*	Input MC_cluster: the clustering struct
*	Input delta: the max length of the move (in radians around the COM of the cluster)
*	Input: mspin_cluster: Function that spins a cluster in the current geometry
*/


int CMV_spin_mc (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster, double delta,
			double *radius, double psudo_temp, int periodic, double *paramsG,
			double (*potential)(double,double*),
			void (*spin_cluster)(Mc_state*,Mc_cell*,Mc_clusters*,int,double,double*),
 			double paramsP[]) {
	
	/*select cluster to move*/
	Mc_state *old_state = malloc(sizeof(Mc_state));
	Mc_cell *old_cell   = malloc(sizeof(Mc_cell));
	MCS_state_clone(old_state,state);
	CLL_cell_clone(old_cell,cell,state->n_particles);

	int head = cluster->head_of[randomMT() % state->n_particles];

	double old_energy = 0;
	int n = head;
	while ( n != -1) {
		old_energy += CMV_energy_calc(state,cell,cluster, state->x[n], state->y[n], 
					      state->z[n], periodic, paramsG, n, 
					      cell->of_particle[n], psudo_temp,potential,paramsP);
		n = cluster->ll[n];
	}

	spin_cluster(state,cell,cluster,head,delta,radius);

	double energy = 0;

	n = head;
	while ( n != -1) {
		energy += CMV_energy_calc(state,cell,cluster, state->x[n], state->y[n], 
					      state->z[n], periodic, paramsG, n, 
					      cell->of_particle[n], psudo_temp,potential,paramsP);
		n = cluster->ll[n];
	}	
	
	//accepts_reject
	if (randomF() < exp (-1 * (energy-old_energy))) {
		CLL_end(old_cell);
		MCS_end(old_state);
		free(old_cell);
		free(old_state);
		return 1;
	} else {
		MCS_state_deepCopy(state,old_state);
		CLL_cell_deepCopy(cell,old_cell,state->n_particles,state->n_particles);
		CLL_end(old_cell);
		MCS_end(old_state);
		free(old_cell);
		free(old_state);
		return 0;
	}
}


/* FUNCTION: move_cluster takes a whole cluster and moves it as a whole. The new
 * position of the centre of the mass of the cluster moves arounnd the sphere
 * by an angle less than delta_cluster.
*  Input head: the head of the cluster to be moved
* 	Input: Mc_state: the current state of the system
*	Input: Mc_cell: the cell list for the system
*	Input MC_cluster: the clustering struct
*	Input delta: the max length of the move (an radians aroud the sphere)
*	Input: radius: the radius of the sphere
*/


void CMV_move_cluster_sphere (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster, int head, 
	double delta,double *paramsG) {

	double x_prime,y_prime,z_prime;
	int n;
	/*get random axis*/
	double u = 2*randomF() - 1;
	double phi = 2*M_PI*randomF();
	double x_ax = sqrt(1-u*u) * cos(phi);
	double y_ax = sqrt(1-u*u) * sin(phi);
	double z_ax = u;

	/*get roatation amount*/
	double move_angle = delta * randomF();
	double cos_angle = cos(move_angle);
	double sin_angle = sin(move_angle);

	/*get matrix*/	
	double xx = x_ax*x_ax + (1-x_ax*x_ax)*cos_angle;
	double xy = x_ax*y_ax*(1-cos_angle) - z_ax*sin_angle;
	double xz = x_ax*z_ax*(1-cos_angle) + y_ax*sin_angle;
	double yx = x_ax*y_ax*(1-cos_angle) + z_ax*sin_angle;
	double yy = y_ax*y_ax + (1-y_ax*y_ax)*cos_angle;
	double yz = y_ax*z_ax*(1-cos_angle) - x_ax*sin_angle;
	double zx = x_ax*z_ax*(1-cos_angle) - y_ax*sin_angle;
	double zy = y_ax*z_ax*(1-cos_angle) + x_ax*sin_angle;
	double zz = z_ax*z_ax + (1-z_ax*z_ax)*cos_angle;

	n = head;
	while (n != -1) {
		x_prime = xx*state->x[n] + xy*state->y[n] + xz*state->z[n];
		y_prime = yx*state->x[n] + yy*state->y[n] + yz*state->z[n];
		z_prime = zx*state->x[n] + zy*state->y[n] + zz*state->z[n];
		state->x[n] = x_prime;
		state->y[n] = y_prime;
		state->z[n] = z_prime;
		CLL_update_particle(cell, state, n);
		n = cluster->ll[n];	
	}
}

/* FUNCTION: spin_cluster takes a whole cluster and spins it as a whole. The centre of mass does not change
*  Input head: the head of the cluster to be moved
* 	Input: Mc_state: the current state of the system
*	Input: Mc_cell: the cell list for the system
*	Input MC_cluster: the clustering struct
*	Input delta: the max length of the move (in radians around the COM of the cluster)
*/


void CMV_spin_cluster_sphere (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster, int head, 
				double delta,double *paramsG) {

	int n_in_cluster = 0;
	int n;
	double x_ax = 0, y_ax = 0, z_ax = 0;
	double x_prime,y_prime,z_prime;

	/*get centre of mass*/
	n = head;
	while (n != -1) {
		x_ax += state->x[n];
		y_ax += state->y[n];
		z_ax += state->z[n];

		n_in_cluster++;
		n = cluster->ll[n];	
	}

	x_ax = x_ax/n_in_cluster;
	y_ax = y_ax/n_in_cluster;
	z_ax = z_ax/n_in_cluster;

	/*project onto unit sphere*/
	double length = sqrt(x_ax*x_ax + y_ax*y_ax + z_ax*z_ax);
	x_ax *= (1.0/length);
	y_ax *= (1.0/length);
	z_ax *= (1.0/length);

	/*get roatation amount*/
	double move_angle = delta * randomF();
	double cos_angle = cos(move_angle);
	double sin_angle = sin(move_angle);

	/*get matrix*/	
	double xx = x_ax*x_ax + (1-x_ax*x_ax)*cos_angle;
	double xy = x_ax*y_ax*(1-cos_angle) - z_ax*sin_angle;
	double xz = x_ax*z_ax*(1-cos_angle) + y_ax*sin_angle;
	double yx = x_ax*y_ax*(1-cos_angle) + z_ax*sin_angle;
	double yy = y_ax*y_ax + (1-y_ax*y_ax)*cos_angle;
	double yz = y_ax*z_ax*(1-cos_angle) - x_ax*sin_angle;
	double zx = x_ax*z_ax*(1-cos_angle) - y_ax*sin_angle;
	double zy = y_ax*z_ax*(1-cos_angle) + x_ax*sin_angle;
	double zz = z_ax*z_ax + (1-z_ax*z_ax)*cos_angle;

	n = head;
	while (n != -1) {
		x_prime = xx*state->x[n] + xy*state->y[n] + xz*state->z[n];
		y_prime = yx*state->x[n] + yy*state->y[n] + yz*state->z[n];
		z_prime = zx*state->x[n] + zy*state->y[n] + zz*state->z[n];
		state->x[n] = x_prime;
		state->y[n] = y_prime;
		state->z[n] = z_prime;
		CLL_update_particle(cell, state, n);
		n = cluster->ll[n];	
	}
}

/* FUNCTION: move_cluster takes a whole cluster and moves it as a whole. The new
 * position of the centre of the mass of the cluster moves arounnd the plane
 * by a distance less than delta_cluster.
*  Input head: the head of the cluster to be moved
* 	Input: Mc_state: the current state of the system
*	Input: Mc_cell: the cell list for the system
*	Input MC_cluster: the clustering struct
*	Input delta: the max length of the move (an radians aroud the sphere)
*	Input: paramsG: the box size of the plane
*/

void CMV_move_cluster_plane (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster,int head, 
		double delta,double *paramsG) {

	if (state->n_particles == 0) return;

	double x_move = (2*randomF()-1)*delta;
	double y_move = (2*randomF()-1)*delta;

	int n = head;
	while (n != -1) {
		state->x[n] += x_move;
		state->y[n] += y_move;

		if (state->x[n] > paramsG[0]/2.0) {
			state->x[n] -= paramsG[0];
		} else if (state->x[n] < -1*paramsG[0]/2.0) {
			state->x[n] += paramsG[0];
		}

		if (state->y[n] > paramsG[1]/2.0) {
			state->y[n] -= paramsG[1];
		} else if (state->y[n] < -1*paramsG[1]/2.0) {
			state->y[n] += paramsG[1];
		}
		CLL_update_particle(cell, state, n);
		n = cluster->ll[n];
	}

}

/* FUNCTION: spin_cluster takes a whole cluster and spins it as a whole. The centre of mass does not change
*  Input head: the head of the cluster to be moved
* 	Input: Mc_state: the current state of the system
*	Input: Mc_cell: the cell list for the system
*	Input MC_cluster: the clustering struct
*	Input delta: the max length of the move (in radians around the COM of the cluster)
*/


void CMV_spin_cluster_plane (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster,int head, 
		double delta,double *paramsG) {

	if (state->n_particles == 0) return;

	double x_ax = 0, y_ax = 0;
	int n_in_cluster;
	/*get centre of mass*/
	int n = head;
	while (n != -1) {
		x_ax += state->x[n];
		y_ax += state->y[n];

		n_in_cluster++;
		n = cluster->ll[n];	
	}

	x_ax = x_ax/n_in_cluster;
	y_ax = y_ax/n_in_cluster;

	double angle = (2*randomF()-1)*delta;

	double cos_angle = cos(angle);
	double sin_angle = sin(angle);

	n = head;
	while (n != -1) {
		state->x[n] = cos_angle*(state->x[n] - x_ax) - sin_angle*(state->y[n] - y_ax) + x_ax;
		state->y[n] = sin_angle*(state->x[n] - x_ax) + cos_angle*(state->y[n] - y_ax) + y_ax;

		if (state->x[n] > paramsG[0]/2.0) {
			state->x[n] -= paramsG[0];
		} else if (state->x[n] < -1*paramsG[0]/2.0) {
			state->x[n] += paramsG[0];
		}

		if (state->y[n] > paramsG[1]/2.0) {
			state->y[n] -= paramsG[1];
		} else if (state->y[n] < -1*paramsG[1]/2.0) {
			state->y[n] += paramsG[1];
		}
		CLL_update_particle(cell, state, n);
		n = cluster->ll[n];
	}

}


/*----------------------------------END OF PUBLIC FUNCTIONS------------------------------*/

/*helper function for get_clusters*/
void CMV_connect (Mc_state *state, Mc_cell *cells, Mc_clusters *cluster,  
			int i, int j, double psudo_temp, int periodic, double *paramsG,
			double (*potential)(double,double*), double paramsP[]) {

	int n;
	double energy = potential ( GEO_distance (state->x[i],state->y[i],state->z[i],
							state->x[j], state->y[j], state->z[j],
							periodic, paramsG), 
				     paramsP);
	double r = randomF();

	if (r < 1-exp(energy/psudo_temp)) {
		if (cluster->head_of [i] != cluster->head_of[j]) {
			/*attach head of i to tail of j*/
			n = i;
			while(cluster->ll[n] != -1) {
				n = cluster->ll[n];
			}
			cluster->ll[n] = cluster->head_of[j];
			n = cluster->head_of[j];
			while(n != -1) {
				cluster->head_of[n] = cluster->head_of[i];
				n = cluster->ll[n];
			}
		}	
	}	
}

double CMV_energy_calc (Mc_state *state, Mc_cell *cell, Mc_clusters *cluster,
		        double x, double y, double z, int periodic, double *paramsG,
			int index, int cell_of, double psudo_temp,
			double (*potential)(double,double*), double paramsP[]) {

	double energy = 0;
	int n;
	int current_cell = cell_of;
	int head_of_index = cluster->head_of[index];
	double x_i = state->x[index];
	double y_i = state->y[index];
	double z_i = state->z[index];
	double potentialV;

	n = cell->head_of [current_cell];
	while (n != index && n != -1) {
		if (cluster->head_of[n] != head_of_index){
			potentialV = potential(GEO_distance (state->x[n],state->y[n],state->z[n],
							     x_i, y_i, z_i,periodic,paramsG),
						paramsP);
			energy += potentialV - fmin(potentialV/psudo_temp,0.0);
		}
		n = cell->ll[n];
			 
	}
	if (n != -1) {
		n = cell->ll[n];
	}
	while (n != -1) {
		if (cluster->head_of[n] != head_of_index){
			potentialV = potential(GEO_distance (state->x[n],state->y[n],state->z[n],
							     x_i, y_i, z_i,periodic,paramsG),
						     paramsP);
			energy += potentialV - fmin(potentialV/psudo_temp,0.0);
		}
		n = cell->ll[n];	
	}
	int j;
	for (j = 0; j < 26; j++) {
		if (cell->neighbours[cell_of][j] == -1) break;
		current_cell = cell->neighbours[cell_of][j];
		n = cell->head_of[current_cell];
		while (n != -1) {
			if (cluster->head_of[n] != head_of_index){
				potentialV = potential(GEO_distance (state->x[n],state->y[n],
							     state->z[n], x_i, y_i,
							      z_i,periodic,paramsG),
							  paramsP);
				energy += potentialV - fmin(potentialV/psudo_temp,0.0);
			}
			n = cell->ll[n];
		}
	}
	return energy;
}

