#include"std_h.h"
#include"mc_math_h.h"
#include"lennard_jones_ts_lambda_h.h"

/* a module for implementing the truncated, shifted and smoothed Lennard-Jones potential
 * modified with a switching parameter for thermodynamic integration.
 * The prefix is ldtsl_
 */

/*FUNCTION: get params: A helper function to set up the params array for this function to be
 * passed. Not required but useful, as it does all the differentiating for you etc.
 *
 * Output: double params[] (array of size 5)
 *		params[0] = epsilon, the inverse temperature in reduced units
 *		params[1] = The cutoff distance for the potential. The unit of distance is to
 *			the distance at which the Lennard Jones potential
 *			has it's minimum.
 *		params[2] = the value of the Lennard jones potential at the cutoff
 *		params[3] = the value of the differential of the potential
 *				at the cutoff
 *		params[4] = the inverse magnitude of the potential at r = 1, eps = 1
 *		params[5] = the value of lambda
 *		params[6] = the value of alpha
 *		params[7] = the value of dv/d lambda at the cutoff
 *		params[8] = the value of d/d lambda dv/dr at the cutoff
 *
 * Input: double epsilon: The value of the inverse temperature
 * input double cutoff: the cutoff length. 	
 */

void ldtsl_get_params(double params[],double epsilon,double cutoff,double lambda,double alpha) {
	params[0] = epsilon;
	params[1] = cutoff;
	params[5] = lambda;
	params[6] = alpha;
	double r6 = 1/(alpha*(1-lambda)+pow(cutoff,6));
	params[2] = epsilon*(r6*r6 - 2*r6);
	params[3] = -12*epsilon*pow(cutoff,5)*(r6*r6*r6-r6*r6);
	params[7] = 2*epsilon*alpha*(r6*r6*r6-r6*r6);
	params[8] = -12*epsilon*alpha*pow(cutoff,5)*(3*r6*r6*r6*r6 - 2*r6*r6*r6);
	params[4] = (1-params[2] + (1-params[1])*params[3]);
}

/*
 FUNCTION: double lennard_jones: calculates the pairwise ld potential
 * OUTPUT: ld potential in units of kT
 * INPUT: double distance the separation of the two particles
 */

double ldtsl_potential(double distance, double params[]) {
	if(distance > params[1]) {
		return 0.0;
	}
	else {
		double r6 = 1/(params[6]*(1-params[5])+pow(distance,6));
		return params[5]*(params[0]*(r6*r6 - 2*r6) - params[2] -
                                    (distance-params[1])*params[3]);
	}

}

/*
 FUNCTION: double d_lennard_jones: calculates the pairwise ld potential differentiated by lambda for
 * the purpose of the thermodynamic integration.
 * OUTPUT: ld potential in units of kT
 * INPUT: double distance the separation of the two particles
 */

double ldtsl_d_potential(double distance, double params[]) {
	if(distance > params[1]) {
		return  0.0;
	}
	else {
		double r6 = 1/(params[6]*(1-params[5])+pow(distance,6));
		return  params[0]*(r6*r6 - 2*r6) - params[2] - (distance-params[1])*params[3] + 
			params[5]*(params[0]*2.0*params[6]*(r6*r6*r6-r6*r6) - params[7] - 
                                  (distance-params[1])*params[8]);
	}	
}






