#ifndef MONTE_CARLO_H 
#define MONTE_CARLO_H

/* A module that contains functions for performig monte carlo moves in two dimensions,
 * utilising a cell list which exists in three dimensions around the surface.
 * It comes with a few functions for getting basic output date (frames, nearest neighbours, saving).
 * 
 *
 * The prefix is mc_
 */

/*Function: Initialise. This function places particles in a system State. 
 *      If READFILE is no, then N particles (N read from Input struct) will be 
 *	placed into the system at random.
 *
 *	If READFILE is yes, then the state will be loaded from files "final.csv"
 *	into the state struct. The values in range are assumed to be correct and not checked.
 *
 *
 * input int readFile: 1 if file is to be read, 0 otherwise
 * int n_particles: The number of particles required
 * random: gets a random, well sampled position on the surface of interest
 */

void mc_initialise (Mc_state *state, int readFile, int restart, int n_particles, double *params, 
			char run_id[], void (*random)(double *, double *, double *,double *),
			double *deltaR, double *deltaC, double *deltaS);


/*Function: particle_move. Performs a full single particle move on one particle in the state
 *
 * Output: 1 if the move is accepted, 0 otherwise
 * Param: State: the current state mc_state struct for the system (to be updated)
 * Param: cell: the current (up to date) cell list for the system (to be updated)
 * input: 
 *	double(*potential) a function which returns the potential felt by two particles
 *	separated by a set distance (the first argument).
 *	(*move) a function that suggest random moves for the particle with coords that are the 
 *	first three arguments
 *	paramsP is the parameter list for the user-defined potential function
 *	paramsD is some paramters to pss to the move function for the move function
 *		as is delta_r
 **/

int mc_particle_move (Mc_state *state, Mc_cell *cell,int periodic, double delta_r,
			double (*potential)(double,double*),
			void (*move) (double*,double*,double*,double*,double),
			double paramsP[], double paramsD[]);



/*
 * Function: particle_add. Performs a grand cannonical particle addition step
 * Output: 1 if the move is accepted, 0 otherwise
 * Param: State: the current state mc_state struct for the system (to be updated)
 * Param: cell: the current (up to date) cell list for the system (to be updated)
 *
 *
 * input: 
 *	double(*potential) a function which returns the potential felt by two particles
 *	separated by a set distance (the first argument).
 * 	random: gets a random, well sampled position on the surface of interest
 *	(*move) a function that suggest random moves for the particle with coords that are the 
 *	paramsP is the parameter list for the user-defined potential function
 *	paramsD is the parameter list for the user-defined geometry function
 *	activity is the activity of the gc ensemble
 */


int mc_particle_add (Mc_state *state, Mc_cell *cell, int periodic,
			double (*potential)(double,double*),
			void (*random)(double *, double *, double *,double *),
			double paramsP[], double paramsD[], double activity);

/*
 * Function: particle_remove. Performs a grand cannonical particle removal step
 * Output: 1 if the move is accepted, 0 otherwise
 * Param: State: the current state mc_state struct for the system (to be updated)
 * Param: cell: the current (up to date) cell list for the system (to be updated)
 *
 *
 * input: 
 *	double(*potential) a function which returns the potential felt by two particles
 *	separated by a set distance (the first argument).
 *	(*move) a function that suggest random moves for the particle with coords that are the 
 *	paramsP is the parameter list for the user-defined potential function
 *	paramsD is the parameter list for the user-defined geometry function
 *	activity is the activity of the gc ensemble
 */

int mc_particle_remove (Mc_state *state, Mc_cell *cell, int periodic,
			double (*potential)(double,double*),
			double paramsP[], double paramsD[], double activity);


/*
 * Function mc_Record: Spits out a frame and attaches it to the end of history.csv
 * Input: State: the current state of the system
 */

void mc_record(Mc_state *state);

/*
 *Function mc_record_neighbours: spits out the value of some order parameter function for each
 * particle in a way that can be used by visulaisation programs
 */

void mc_record_order( double *orders, int n, char *name );

/*Function: MC_save: saves the current state of the system: can be used for restarts*/
void mc_save(Mc_state *state, char run_id[], double delta_r, double deltaC, double deltaS) ;

#endif
