/*this module simply implemetns the basics of umbrella sampling, without any assumptions about the
 * type of order parameter. Assumes no  replica exchange.
 *
 * the module prefix is _UBB*/

/* FUNCTION: constructor: makes a new histogram big enough to contain all posible nuclear sizes
 * 
 * Input: histogram: the histogram to be initialised
 * Input: Volume: the volume of the system in units of particle volume
 * Input: *name: the name of the histogram
 */

#include"mc_state_h.h"
#include"cell_list_h.h"
#include"umbrella_basic_h.h"
#include"std_h.h"
#include"mc_math_h.h"
#include"twister_h.h"

void UBB_new_histogram(Mc_umbrellaB *histogram, double volume, char *name, int restart) {

	int saturation = (int) (volume/(0.25*M_PI));
	histogram->max_size = (int) 1.5*saturation;
	histogram->isScaled = 0;
	histogram->freq = malloc(sizeof(long)*histogram->max_size);
	histogram->scaled = malloc(sizeof(double)*histogram->max_size);
strcpy(histogram->name,name);
	int i;

	if (restart) {

		char command[50];
		sprintf(command,"%s_frequencies.csv",histogram->name);
		FILE *fp;
		fp = fopen(command,"r");
		for (i = 0; i < histogram->max_size; i++) {
			fscanf(fp,"%*d %*c %ld %*[\n]",(histogram->freq)+i);
			histogram->scaled[i] = 0.0;
		}
		fclose(fp);

	} else {
		for(i = 0; i < histogram->max_size; i++) {
			histogram->freq[i] = 0;
			histogram->scaled[i] = 0.0;
		}
	}
}

/*Function add value: add a new nucleus to the histogram
 *
 * Input: histogram: the histogram to be initialised
 * Input: value: the value to be added
 **/
void UBB_add_value(Mc_umbrellaB *histogram, int value) {
	histogram->isScaled = 0;

	if (value < histogram->max_size) {
		histogram->freq[value]++;
	}
}

/* Function: scale:  scales the histogram acording to the rules of umbrella sampling
 *
 * Input: Mc_umbrella : the histogram to be scaled
 * Input: lambda: the spring constant of the umbrella potential
 *input : n_target: The target nuclear size
 */

void UBB_scale(Mc_umbrellaB *histogram, double lambda, int n_target) {
	histogram->isScaled = 1;
	int i;
	for (i = 0; i < histogram->max_size; i++) { 
		histogram->scaled[i] = (double) histogram->freq[i] / 
					        (exp(-0.5*lambda*(i-n_target)*(i-n_target)) );
	}
}

/* Function: print: prints out both the raw feqecncy and scaled histogram data
 *
 * Input histogramL the histgoram to be printed
 **/

void UBB_print(Mc_umbrellaB *histogram) {
	int i;
	FILE *fp;
	char command[50];

	sprintf(command,"%s_frequencies.csv",histogram->name);
	fp =fopen(command,"w");
	for (i = 0; i < histogram->max_size; i++) {
		fprintf(fp,"%d,%ld\n",i,histogram->freq[i]);
	}

	fclose(fp);	

	sprintf(command,"%s_scaled.csv",histogram->name);
	fp = fopen(command,"w");

	if (histogram->isScaled == 0) {
		fprintf(fp,"Scaled histogram is not up to date with frequencies\n");
	}

	for (i = 0 ; i < histogram->max_size; i++) {
		fprintf(fp,"%d,%lf\n",i,histogram->scaled[i]);
	}
	fclose(fp);
}

/*Destructor*/
void UBB_destroy(Mc_umbrellaB *histogram) {
	free(histogram->freq);
	free(histogram->scaled);
}


/*FUNCTION UBB_move: This function checks whether the latest batch of umbrella moves is accepted
 * or rejected. This is based on the values of the order parater u_rec and u_old
 *
 * Output: 1 if the move is aacpected, 0 otherwise
 * Param: *mc_state state: the current state of the system
 * Param: *mc_state old_state: the old state of the system
 * Output: u_rec, the final nulcear size for this system
 * param: u_old: the previous size of the largest nucleus in the system
 * input: lambda: the weighting of the umbrealla sampling bias energy
 * input n_target: the target nuclear size */

int UBB_move(Mc_state *state, Mc_state *old_state, Mc_cell *cell, Mc_cell *old_cell, int *u_rec,
		int *u_old, double lambda, int n_target) {
	double energy_old = 0.5 * lambda * ((*u_old) - n_target) * ((*u_old) - n_target);
	double energy_new = 0.5 * lambda * ((*u_rec) - n_target) * ((*u_rec) - n_target);
	
	if (randomF() < exp (-1 * (energy_new-energy_old))) {
		*u_old = *u_rec;
		CLL_cell_deepCopy ( old_cell, cell, state->n_particles, old_state->n_particles );
		MCS_state_deepCopy ( old_state, state);
		return 1;
	} else {
		*u_rec = *u_old;
		CLL_cell_deepCopy (cell, old_cell, old_state->n_particles, state->n_particles);
		MCS_state_deepCopy (state, old_state);
		return 0;
	}

}

/*FUNCTION UBB_move_max: This fuction checks whether the latest batch of umbrella moves
 * is accepted or rejected. All moves with nucleus > n_max get rejected, otherwise accpeted.
 *
 *Output: 1 if the move is accpected, zero otherwise
 *Param: *mc_state state: the current state of the system
 *Param: *mc_state old_state: the old state of the system
 *Output: u_rec, the final nulcear size for this system
 *param: u_old: the previous size of the largest nucleus in the system
 * input n_max: the largest allowed nucleus
 */

int UBB_move_max (Mc_state *state, Mc_state *old_state, Mc_cell *cell, Mc_cell *old_cell,
		  int *u_rec, int *u_old,int n_max) {

	if (*u_rec <= n_max) {
		*u_old = *u_rec;
		CLL_cell_deepCopy ( old_cell, cell, state->n_particles, old_state->n_particles );
		MCS_state_deepCopy ( old_state, state);
		return 1;
	} else {
		*u_rec = *u_old;
		CLL_cell_deepCopy (cell, old_cell, old_state->n_particles, state->n_particles);
		MCS_state_deepCopy (state, old_state);
		return 0;
	}

}




